#include <sys/socket.h>
#include <errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ks_utils.h"
#include "log.h"
#include "ini.h"

int main(int argc, char** argv) {

    /* allocate and load config file */
    ini_t* config = ini_load("config.ini");
    
    const char* keycloak_ip = ini_get(config, "Keycloak", "keycloak_ip");
    const char* keycloak_port = ini_get(config, "Keycloak", "keycloak_port");
    const char* kc_client_id = ini_get(config, "Keycloak", "kc_client_id");
    const char* kc_client_secret = ini_get(config, "Keycloak", "kc_client_secret");

    const char* keyserver_ip = ini_get(config, "KeyServer", "keyserver_ip");
    const char* keyserver_port = ini_get(config, "KeyServer", "keyserver_port");

    const char* target_sae_ip = ini_get(config, "SAEs", "target_sae_ip");
    const char* target_sae = ini_get(config, "SAEs", "target_sae");
    const char* this_sae = ini_get(config, "SAEs", "this_sae");

    int socket_port = atoi(ini_get(config, "Socket data", "socket_port"));
    int retry_conn_sec = atoi(ini_get(config, "Socket data", "retry_conn_sec"));

    /* check if log dir exists otherwise create it */
    struct stat st = {0};
    if (stat("log", &st) == -1) {
        mkdir("log", 0700);
    }

    /* log function related code */
    FILE *fLog = NULL;
    FILE *flogTime = NULL;
    fLog = fopen("./log/client_setup.log", "a");
    flogTime = fopen("./log/client_setup.exectime", "a");
    log_add_fp(fLog, LOG_TRACE);
    log_add_fp(flogTime, LOG_TIME);

    /* Vars needed by time functions */
    double w_total, cpu_total;
    w_total = cpu_total = 0;
    struct timespec w_begin, w_end;
    struct timespec cpu_begin, cpu_end;

    int sockfd;
    struct  sockaddr_in servAddr;
    const char message[] = "Sending handle";
    
    struct MemoryStruct p;
    char* acc_token = NULL;
    char buffer[1024] = {0};

    char* myKey = NULL;
    char* myKeyID = NULL;

    FILE* fSec = NULL;

    stopwatch_start(&w_begin, &cpu_begin);

    if(!get_token(&p, keycloak_ip, keycloak_port, kc_client_id, kc_client_secret)) {
        log_debug("RECEIVED: %s", p.memory);
        if(!parse_json_token(p.memory, &acc_token)) {
            log_debug("TOKEN: %s", acc_token);
            free(p.memory);
            p.size = 0;
        }
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Received access token from KeyCloak", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    if(!get_key(&p, acc_token, keyserver_ip, keyserver_port, target_sae, this_sae)) {
        log_debug("RECEIVED: %s", p.memory);
        if(!parse_json_key(p.memory, &myKeyID, &myKey)) {
            log_debug("KEY_ID: %s, KEY: %s", myKeyID, myKey);
            free(p.memory);
            p.size = 0;
        }
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Received key from KeyServer (GET KEY)", &w_total, &cpu_total);

    if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
        log_error("Socket creation error");
        return -1;
    }

    servAddr.sin_family = AF_INET;
    servAddr.sin_port = htons(socket_port);

    // Convert IPv4 and IPv6 addresses from text to binary form
    if(inet_pton(AF_INET, target_sae_ip, &servAddr.sin_addr)<=0) {
        log_error("Invalid address/ Address not supported");
        return -1;
    }

    while (connect(sockfd, (struct sockaddr *)&servAddr, sizeof(servAddr)) < 0) {
        log_error("Connection Failed - Waiting %ds before trying again...", retry_conn_sec);
        sleep(retry_conn_sec);
    }

    stopwatch_start(&w_begin, &cpu_begin);

    // Send KEY HANDLE
    send(sockfd, myKeyID, strlen(myKeyID) , 0);
    log_debug("Key handle sent");
    read(sockfd, buffer, 1024);
    log_debug("%s", buffer);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Sent handle and received ack", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    fSec = fopen("./client_secrets.txt", "w");
    fprintf(fSec, "%s %s", myKeyID, myKey);
    fclose(fSec);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Store key on file", &w_total, &cpu_total);

    free(acc_token);
    free(myKeyID);
    free(myKey);
    ini_free(config);
    close(sockfd);

    log_time("Execution completed: [WALL] %.6f seconds -- [CPU] %.6f seconds\n", w_total, cpu_total);

    char * const exargs[] = {NULL};
    execv("test_client", exargs);
}
