#include "ks_utils.h"

double diff_time(struct timespec* begin, struct timespec* end) {
    long sec = end->tv_sec - begin->tv_sec;
    long nanosec = end->tv_nsec - begin->tv_nsec;
    double elapsed = sec + nanosec*1e-9;
    return elapsed;
}

void stopwatch_start(struct timespec* w_begin, struct timespec* cpu_begin) {

    double w_time, cpu_time;

    clock_gettime(CLOCK_REALTIME, w_begin);
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, cpu_begin);

}

void stopwatch_stop(struct timespec* w_begin, struct timespec* w_end, struct timespec* cpu_begin, struct timespec* cpu_end, 
    char* message, double* w_total, double* cpu_total) {

    double w_time, cpu_time;

    clock_gettime(CLOCK_REALTIME, w_end);
    clock_gettime(CLOCK_PROCESS_CPUTIME_ID, cpu_end);
    w_time = diff_time(w_begin, w_end);
    cpu_time = diff_time(cpu_begin, cpu_end);
    log_time("%s = [WALL] %.6f seconds -- [CPU] %.6f seconds", message, w_time, cpu_time);
    *w_total +=  w_time;
    *cpu_total += cpu_time;
    
}


void sanitize_ks_response(char* s) {
    const char* d = s;
    do {
        while (*d == ' ' || (*d == 'b' && *(d+1) == '\'' )) {
            ++d;
        }
    } while (*s++ = *d++);
}

static size_t write_memory_cb(void* contents, size_t size, size_t nmemb, void* userp) {
    size_t realsize = size * nmemb;
    struct MemoryStruct *mem = (struct MemoryStruct *)userp;

    char *ptr = realloc(mem->memory, mem->size + realsize + 1);
    if(ptr == NULL) {
        /* out of memory! */ 
        log_error("not enough memory (realloc returned NULL)");
        return 0;
    }

    mem->memory = ptr;
    memcpy(&(mem->memory[mem->size]), contents, realsize);
    mem->size += realsize;
    mem->memory[mem->size] = 0;

    return realsize;
}

int get_token(struct MemoryStruct* chunk, const char* keycloak_ip, const char* keycloak_port, const char* keycloak_client_id, const char* keycloak_sec) {

    CURL *curl_handle;
    CURLcode res;

    char* buffer;

    struct curl_slist* slist = NULL;

    chunk->memory = malloc(1);  /* will be grown as needed by the realloc above */ 
    chunk->size = 0;    /* no data at this point */ 

    curl_global_init(CURL_GLOBAL_ALL);

    /* init the curl session */ 
    curl_handle = curl_easy_init();

    /* specify URL to get */
    buffer = (char*)malloc(sizeof(char)*255);
    sprintf(buffer, "http://%s:%s/auth/realms/quantum_auth/protocol/openid-connect/token", keycloak_ip, keycloak_port);
    curl_easy_setopt(curl_handle, CURLOPT_URL, buffer);

    /* Now specify the POST data */
    bzero(buffer, strlen(buffer));
    sprintf(buffer, "client_id=%s&client_secret=%s&grant_type=client_credentials", keycloak_client_id, keycloak_sec);
    curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, buffer);
    
    /* insert custom header -> {'Content-Type': 'application/x-www-form-urlencoded'} */
    slist = curl_slist_append(slist, "Content-Type: application/x-www-form-urlencoded");
    curl_easy_setopt(curl_handle, CURLOPT_HTTPHEADER, slist);

    /* send all data to this function  */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, write_memory_cb);
    
    /* we pass our 'chunk' struct to the callback function */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)chunk);
    
    /* some servers don't like requests that are made without a user-agent
        field, so we provide one */ 
    curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

    /* get it! */ 
    res = curl_easy_perform(curl_handle);

    /* check for errors */ 
    if(res != CURLE_OK) {
    fprintf(stderr, "curl_easy_perform() failed: %s\n",
            curl_easy_strerror(res));
    }
    else {
    /*
        * Now, our chunk.memory points to a memory block that is chunk.size
        * bytes big and contains the remote file.
        *
        * Do something nice with it!
        */ 

    log_debug("%lu bytes retrieved", (unsigned long)chunk->size);
    }

    /* free custom headers */
    curl_slist_free_all(slist);

    /* cleanup curl stuff */ 
    curl_easy_cleanup(curl_handle);

    /* we're done with libcurl, so clean it up */ 
    curl_global_cleanup();

    free(buffer);

    return 0;
}

int get_key(struct MemoryStruct* chunk, const char* token, const char* key_server_ip, const char* key_server_port, const char* target_sae, const char* this_sae_str) {
    
    CURL *curl_handle;
    CURLcode res;

    char* header;
    char* buffer;

    struct curl_slist* slist = NULL;
    struct json_object* post_data = NULL;
    struct json_object* number = NULL;
    struct json_object* klen = NULL;
    struct json_object* this_sae = NULL;

    chunk->memory = malloc(1);  /* will be grown as needed by the realloc above */ 
    chunk->size = 0;    /* no data at this point */ 

    curl_global_init(CURL_GLOBAL_ALL);

    /* init the curl session */ 
    curl_handle = curl_easy_init();

    /* specify URL to get */
    buffer = (char*)malloc(sizeof(char)*255);
    sprintf(buffer, "http://%s:%s/api/v1/keys/%s/enc_keys", key_server_ip, key_server_port, target_sae);
    curl_easy_setopt(curl_handle, CURLOPT_URL, buffer);

    /* Now specify the POST data */

    post_data = json_object_new_object();
    number = json_object_new_int(1);
    klen = json_object_new_int(128);
    this_sae = json_object_new_string(this_sae_str);
    json_object_object_add(post_data, "number", number);
    json_object_object_add(post_data, "size", klen);
    json_object_object_add(post_data, "SAE_ID", this_sae);

    curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, json_object_to_json_string(post_data));

    /* insert custom header -> {'Authorization': 'Bearer ' + token} */
    header = (char*)malloc(sizeof(char) * (strlen(token) + strlen("Authorization: Bearer ") + 1));
    sprintf(header, "Authorization: Bearer %s", token);
    slist = curl_slist_append(slist, header);
    slist = curl_slist_append(slist, "Content-type: ");
    free(header);
    curl_easy_setopt(curl_handle, CURLOPT_HTTPHEADER, slist);

    /* send all data to this function  */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, write_memory_cb);
    
    /* we pass our 'chunk' struct to the callback function */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)chunk);
    
    /* some servers don't like requests that are made without a user-agent
        field, so we provide one */ 
    curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

    /* get it! */ 
    res = curl_easy_perform(curl_handle);

    /* check for errors */ 
    if(res != CURLE_OK) {
    fprintf(stderr, "curl_easy_perform() failed: %s\n",
            curl_easy_strerror(res));
    }
    else {
    /*
        * Now, our chunk.memory points to a memory block that is chunk.size
        * bytes big and contains the remote file.
        *
        * Do something nice with it!
        */ 

    log_debug("%lu bytes retrieved", (unsigned long)chunk->size);
    }

    /* free custom headers */
    curl_slist_free_all(slist);

    /* cleanup curl stuff */ 
    curl_easy_cleanup(curl_handle);

    /* we're done with libcurl, so clean it up */ 
    curl_global_cleanup();

    json_object_put(post_data);

    free(buffer);

    return 0;
}

int get_key_with_id(struct MemoryStruct* chunk, const char* token, const char* key_id, const char* key_server_ip, 
    const char* key_server_port, const char* target_sae){

    CURL *curl_handle;
    CURLcode res;

    char* header;
    char* buffer;

    struct curl_slist* slist = NULL;
    struct json_object* post_data = NULL;
    struct json_object* post_data_arr = NULL;
    struct json_object* key_id_str = NULL;
    struct json_object* key_id_obj = NULL;

    chunk->memory = malloc(1);  /* will be grown as needed by the realloc above */ 
    chunk->size = 0;    /* no data at this point */ 

    curl_global_init(CURL_GLOBAL_ALL);

    /* init the curl session */ 
    curl_handle = curl_easy_init();

    /* specify URL to get */
    buffer = (char*)malloc(sizeof(char)*255);
    sprintf(buffer, "http://%s:%s/api/v1/keys/%s/dec_keys", key_server_ip, key_server_port, target_sae);
    curl_easy_setopt(curl_handle, CURLOPT_URL, buffer);

    /* Now specify the POST data */

    key_id_obj = json_object_new_object();
    post_data = json_object_new_object();
    key_id_str = json_object_new_string(key_id);
    json_object_object_add(key_id_obj, "key_ID", key_id_str);
    post_data_arr = json_object_new_array();
    json_object_array_add(post_data_arr, key_id_obj);
    json_object_object_add(post_data, "key_IDs", post_data_arr);

    curl_easy_setopt(curl_handle, CURLOPT_POSTFIELDS, json_object_to_json_string(post_data));
    log_debug("SENDING: %s", json_object_to_json_string(post_data));

    /* insert custom header -> {'Authorization': 'Bearer ' + token} */
    header = (char*)malloc(sizeof(char) * (strlen(token) + strlen("Authorization: Bearer ") + 1));
    sprintf(header, "Authorization: Bearer %s", token);
    slist = curl_slist_append(slist, header);
    slist = curl_slist_append(slist, "Content-type: ");
    free(header);
    curl_easy_setopt(curl_handle, CURLOPT_HTTPHEADER, slist);

    /* send all data to this function  */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEFUNCTION, write_memory_cb);

    /* we pass our 'chunk' struct to the callback function */ 
    curl_easy_setopt(curl_handle, CURLOPT_WRITEDATA, (void *)chunk);

    /* some servers don't like requests that are made without a user-agent
        field, so we provide one */ 
    curl_easy_setopt(curl_handle, CURLOPT_USERAGENT, "libcurl-agent/1.0");

    /* get it! */ 
    res = curl_easy_perform(curl_handle);

    /* check for errors */ 
    if(res != CURLE_OK) {
    fprintf(stderr, "curl_easy_perform() failed: %s\n",
        curl_easy_strerror(res));
    }
    else {
    /*
    * Now, our chunk.memory points to a memory block that is chunk.size
    * bytes big and contains the remote file.
    *
    * Do something nice with it!
    */ 

    log_debug("%lu bytes retrieved", (unsigned long)chunk->size);
    }

    /* free custom headers */
    curl_slist_free_all(slist);

    /* cleanup curl stuff */ 
    curl_easy_cleanup(curl_handle);

    /* we're done with libcurl, so clean it up */ 
    curl_global_cleanup();

    json_object_put(post_data);

    free(buffer);

    return 0;
}

int parse_json_token(char* buffer, char** token_str) {
    
    struct json_object* parsed_json;
    struct json_object* token;
    int token_sz = 0;

    parsed_json = json_tokener_parse(buffer);
    json_object_object_get_ex(parsed_json, "access_token", &token);
    token_sz = json_object_get_string_len(token);
    if (!token_sz)
        return 1;
    *token_str = (char*)malloc(sizeof(char)*(token_sz + 1));
    strncpy(*token_str, json_object_get_string(token), token_sz);
    (*token_str)[token_sz] = '\0';
    json_object_put(parsed_json);
    return 0;
}

int parse_json_key(char* buffer, char** key_id_str, char** key_str) {
    
    char *sanitized_buf = (char*)malloc(sizeof(char)*strlen(buffer));
    sscanf(buffer, "\"[%[^]]\"", sanitized_buf);
    sanitize_ks_response(sanitized_buf);

    struct json_object* parsed_json;
    struct json_object* key_id;
    struct json_object* key;

    int key_id_sz, key_sz;
    key_id_sz = key_sz = 0;

    parsed_json = json_tokener_parse(sanitized_buf);
    json_object_object_get_ex(parsed_json, "key_ID", &key_id);
    json_object_object_get_ex(parsed_json, "key", &key);
    key_id_sz = json_object_get_string_len(key_id);
    key_sz = json_object_get_string_len(key);

    if (!key_id_sz || !key_sz)
    return 1;

    *key_id_str = (char*)malloc(sizeof(char)*key_id_sz + 1);
    *key_str = (char*)malloc(sizeof(char)*key_sz + 1);

    strncpy(*key_id_str, json_object_get_string(key_id), key_id_sz);
    strncpy(*key_str, json_object_get_string(key), key_sz);

    (*key_id_str)[key_id_sz] = '\0';
    (*key_str)[key_sz] = '\0';

    json_object_put(parsed_json);
    free(sanitized_buf);

    return 0;
}

int parse_json_key_id(char* buffer, char** key_id_str, char** key_str) {
    
    sanitize_ks_response(buffer);

    struct json_object* sanitized_pjson_array;
    struct json_object* parsed_json_array;
    struct json_object* parsed_json;
    struct json_object* key_id;
    struct json_object* key;
    struct json_object* key_obj;

    int key_id_sz, key_sz;
    key_id_sz = key_sz = 0;

    parsed_json = json_tokener_parse(buffer);
    json_object_object_get_ex(parsed_json, "keys", &parsed_json_array);
    sanitized_pjson_array = json_tokener_parse(json_object_get_string(parsed_json_array));
    key_obj = json_object_array_get_idx(sanitized_pjson_array, 0);
    json_object_object_get_ex(key_obj, "key_ID", &key_id);
    json_object_object_get_ex(key_obj, "key", &key);
    key_id_sz = json_object_get_string_len(key_id);
    key_sz = json_object_get_string_len(key);
    //printf("RECEIVED: %s AND %s\n", json_object_get_string(key_id), json_object_get_string(key));

    if (!key_id_sz || !key_sz)
    return 1;

    *key_id_str = (char*)malloc(sizeof(char)*(key_id_sz + 1));
    *key_str = (char*)malloc(sizeof(char)*(key_sz + 1));

    *key_id_str = strncpy(*key_id_str, json_object_get_string(key_id), key_id_sz);
    *key_str = strncpy(*key_str, json_object_get_string(key), key_sz);

    (*key_id_str)[key_id_sz] = '\0';
    (*key_str)[key_sz] = '\0';

    int len = strlen(*key_str);

    json_object_put(parsed_json);

    return 0;
}

