#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "ks_utils.h"
#include "ini.h"

int main(int argc, char** argv) {

    /* allocate and load config file */
    ini_t* config = ini_load("config.ini");
    
    const char* keycloak_ip = ini_get(config, "Keycloak", "keycloak_ip");
    const char* keycloak_port = ini_get(config, "Keycloak", "keycloak_port");
    const char* kc_client_id = ini_get(config, "Keycloak", "kc_client_id");
    const char* kc_client_secret = ini_get(config, "Keycloak", "kc_client_secret");

    const char* keyserver_ip = ini_get(config, "KeyServer", "keyserver_ip");
    const char* keyserver_port = ini_get(config, "KeyServer", "keyserver_port");

    const char* target_sae = ini_get(config, "SAEs", "target_sae");
    int socket_port = atoi(ini_get(config, "Socket data", "socket_port"));

    /* check if log dir exists otherwise create it */
    struct stat st = {0};
    if (stat("log", &st) == -1) {
        mkdir("log", 0700);
    }

    /* log function related code */
    FILE *fLog = NULL;
    FILE *flogTime = NULL;
    fLog = fopen("./log/server_setup.log", "a");
    flogTime = fopen("./log/server_setup.exectime", "a");
    log_add_fp(fLog, LOG_TRACE);
    log_add_fp(flogTime, LOG_TIME);

    /* Vars needed by time functions */
    double w_total, cpu_total;
    w_total = cpu_total = 0;
    struct timespec w_begin, w_end;
    struct timespec cpu_begin, cpu_end;

    stopwatch_start(&w_begin, &cpu_begin);

    int listenfd, connfd;
    char* myKey = NULL;
    char* myKeyID = NULL;
    FILE* fSec = NULL;

    struct MemoryStruct p;
    char* acc_token = NULL;
    struct sockaddr_in address;
    int opt = 1;
    int addrlen = sizeof(address);
    char buffer[1024] = {0};

    if ((listenfd = socket(AF_INET, SOCK_STREAM, 0)) == 0) {
        log_error("socket failed");
        exit(EXIT_FAILURE);
    }

    // Forcefully attaching socket to the port 21021
    if (setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR | SO_REUSEPORT, &opt, sizeof(opt))) {
        log_error("setsockopt");
        exit(EXIT_FAILURE);
    }

    address.sin_family = AF_INET;
    address.sin_addr.s_addr = INADDR_ANY;
    address.sin_port = htons(socket_port);

    // Forcefully attaching socket to the port 21021
    if (bind(listenfd, (struct sockaddr *)&address, sizeof(address))<0) {
        log_error("bind failed");
        exit(EXIT_FAILURE);
    }

    if (listen(listenfd, 3) < 0) {
        log_error("listen");
        exit(EXIT_FAILURE);
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Socket created and bound", &w_total, &cpu_total);

    if ((connfd = accept(listenfd, (struct sockaddr *)&address, (socklen_t*)&addrlen))<0) {
        log_error("accept");
        exit(EXIT_FAILURE);
    }

    stopwatch_start(&w_begin, &cpu_begin);

    read(connfd, buffer, 1024);
    log_debug("%s", buffer);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Received handle from client", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    if(!get_token(&p, keycloak_ip, keycloak_port, kc_client_id, kc_client_secret)) {
        log_debug("RECEIVED: %s", p.memory);
        if(!parse_json_token(p.memory, &acc_token)) {
            log_debug("TOKEN: %s", acc_token);
            free(p.memory);
            p.size = 0;
        }
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Received access token from KeyCloak", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    if(!get_key_with_id(&p, acc_token, buffer, keyserver_ip, keyserver_port, target_sae)) {

        log_debug("RECEIVED: %s", p.memory);
        if(!parse_json_key_id(p.memory, &myKeyID, &myKey)) {
            log_debug("KEY_ID: %s, KEY: %s", myKeyID, myKey);
            free(p.memory);
            p.size = 0;
        }
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Received and parsed key from KeyServer (GET KEY WITH ID)", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    fSec = fopen("./secrets.txt", "w");
    fprintf(fSec, "%s %s", myKeyID, myKey);
    fclose(fSec);    

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Key stored on file", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    send(connfd, "OK", 3, 0);
    log_debug("ACK sent");


    free(acc_token);
    free(myKey);
    free(myKeyID);
    close(connfd);
    close(listenfd);
    ini_free(config);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Sent ACK, freed resources", &w_total, &cpu_total);
    log_time("Execution completed: [WALL] %.6f seconds -- [CPU] %.6f seconds\n", w_total, cpu_total);

    char * const exargs[] = {NULL};
    execv("test_server", exargs);
}
