#include <wolfssl/options.h>
#include <wolfssl/ssl.h>
#include <wolfssl/test.h>
#include <sys/socket.h>
#include <errno.h>

#include "ks_utils.h"
#include "ini.h"

char* myKey = NULL;
char* myKeyID = NULL;

static WC_INLINE void custom_tcp_connect(SOCKET_T* sockfd, const char* ip, word16 port,
                               int udp, int sctp, WOLFSSL* ssl, int retry_sec)
{
    SOCKADDR_IN_T addr;
    build_addr(&addr, ip, port, udp, sctp);
    if (udp) {
        wolfSSL_dtls_set_peer(ssl, &addr, sizeof(addr));
    }
    tcp_socket(sockfd, udp, sctp);

    if (!udp) {
        while (connect(*sockfd, (const struct sockaddr*)&addr, sizeof(addr)) != 0) {
            log_error("TCP connect failed - retransmitting in %d...\n", retry_sec);
            sleep(retry_sec);
        }
    }
}

static WC_INLINE unsigned int custom_psk_client_tls13_cb(WOLFSSL* ssl,
        const char* hint, char* identity, unsigned int id_max_len,
        unsigned char* key, unsigned int key_max_len, const char** ciphersuite) {
    int i;
    int b = 0x01;
    const char* userCipher = (const char*)wolfSSL_get_psk_callback_ctx(ssl);

    (void)ssl;
    (void)hint;
    (void)key_max_len;

    XSTRNCPY(identity, myKeyID, strlen(myKeyID));
    XSTRNCPY(key, myKey, strlen(myKey));

    *ciphersuite = userCipher ? userCipher : "TLS13-AES128-GCM-SHA256";

    return 32;   /* length of key in octets or 0 for error */
}

int main(int argc, char** argv) {

    /* allocate and load config file */
    ini_t* config = ini_load("config.ini");

    const char* target_sae_ip = ini_get(config, "SAEs", "target_sae_ip");

    int socket_port = atoi(ini_get(config, "Socket data", "socket_port"));
    int retry_conn_sec = atoi(ini_get(config, "Socket data", "retry_conn_sec"));
    int ntest_msg = atoi(ini_get(config, "Socket data", "ntest_msg"));

    /* log function related code */
    FILE *fLog = NULL;
    FILE *flogTime = NULL;
    fLog = fopen("./log/test_client.log", "a");
    flogTime = fopen("./log/test_client.exectime", "a");
    log_add_fp(fLog, LOG_TRACE);
    log_add_fp(flogTime, LOG_TIME);

    /* Vars needed by time functions */
    double w_total, cpu_total;
    w_total = cpu_total = 0;
    struct timespec w_begin, w_end;
    struct timespec cpu_begin, cpu_end;

    SOCKET_T sockfd;
    WOLFSSL_CTX* ctx;
    WOLFSSL* ssl;
    WOLFSSL_METHOD* method;
    int n;
    char buf[1024];
    struct  sockaddr_in servAddr;
    const char message[] = "Hello, this message is encrypted";
    
    struct MemoryStruct p;
    char* acc_token = NULL;
    char buffer[1024] = {0};
    char readKey[255];
    char readKeyID[255];

    FILE* fSec = NULL;

    stopwatch_start(&w_begin, &cpu_begin);

    fSec = fopen("./client_secrets.txt", "r");
    fscanf(fSec, "%s %s", readKeyID, readKey);
    fclose(fSec);

    myKeyID = (char*)malloc(sizeof(char)*(strlen(readKeyID) + 1));
    myKey = (char*)malloc(sizeof(char)*(strlen(readKey) + 1));

    strncpy(myKeyID, readKeyID, strlen(readKeyID));
    strncpy(myKey, readKey, strlen(readKey));

    myKeyID[strlen(readKeyID)] = '\0';
    myKey[strlen(readKey)] = '\0';

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Retrieved key from file and loaded in memory", &w_total, &cpu_total);

    custom_tcp_connect(&sockfd, target_sae_ip, socket_port, 0, 0, NULL, retry_conn_sec);

    stopwatch_start(&w_begin, &cpu_begin);

    /* initialize wolfssl library */
    wolfSSL_Init();

    method = wolfTLSv1_3_client_method();

    /* make new ssl context */
    if ( (ctx = wolfSSL_CTX_new(method)) == NULL) {
        log_error("wolfSSL_CTX_new error");
    }

    wolfSSL_CTX_set_psk_client_tls13_callback(ctx, custom_psk_client_tls13_cb);

    const char *defaultCipherList = "TLS13-AES128-GCM-SHA256";

    if (wolfSSL_CTX_set_cipher_list(ctx, defaultCipherList) != WOLFSSL_SUCCESS) {
            wolfSSL_CTX_free(ctx); ctx = NULL;
            log_error("client can't set cipher list 2");
    }

    wolfSSL_CTX_set_psk_callback_ctx(ctx, (void*)defaultCipherList);
    
    wolfSSL_CTX_no_dhe_psk(ctx);

    /* make new wolfSSL struct */
    if ( (ssl = wolfSSL_new(ctx)) == NULL) {
        log_error("wolfSSL_new error");
    }

    /* Connect wolfssl to the socket, server, then send message */
    wolfSSL_set_fd(ssl, sockfd);
    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Created wolfssl structures", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);
    wolfSSL_connect(ssl);
    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "TLS Client Hello sent", &w_total, &cpu_total);    


    char* msg = (char*)malloc(sizeof(char)*strlen(message)+2);
    sprintf(msg, "%s:%d", message, 0);


    stopwatch_start(&w_begin, &cpu_begin);
    wolfSSL_write(ssl, msg, strlen(msg));
    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "TLS Handshake completed, first message sent", &w_total, &cpu_total);    

    stopwatch_start(&w_begin, &cpu_begin);
    for(int i = 1; i < ntest_msg; i++) {
        sprintf(msg, "%s:%d", message, i);
        wolfSSL_write(ssl, msg, strlen(msg));
    }

    log_debug("Sent all messages to server");

    while ((n = wolfSSL_read(ssl, buf, (sizeof(buf) -1))) > 0) {
        buf[n] = '\0';
        log_debug("Receiving from server: %s", buf);
	if (buf[n-1] == ntest_msg - 1)
	    break;
    }

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Wrote and read encrypted data", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    /* frees all data before client termination */
    wolfSSL_free(ssl);
    wolfSSL_CTX_free(ctx);
    wolfSSL_Cleanup();

    free(acc_token);
    free(myKeyID);
    free(myKey);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Resources freed", &w_total, &cpu_total);
    log_time("Execution completed: [WALL] %.6f seconds -- [CPU] %.6f seconds\n", w_total, cpu_total);

}

