#include <wolfssl/options.h>
#include <wolfssl/ssl.h>
#include <wolfssl/test.h>
#include <errno.h>

#include "ks_utils.h"
#include "ini.h"

#define MAX_LINE 4096
#define SERV_PORT 21021

char *myKey = NULL;
char *myKeyID = NULL;

static WC_INLINE unsigned int custom_psk_server_tls13_cb(WOLFSSL* ssl,
        const char* identity, unsigned char* key, unsigned int key_max_len,
        const char** ciphersuite)
{
    int i;
    int b = 0x01;
    const char* userCipher = (const char*)wolfSSL_get_psk_callback_ctx(ssl);

    (void)ssl;
    (void)key_max_len;

    if (XSTRNCMP(identity, myKeyID, XSTRLEN(myKeyID)) != 0)
        return 0;
    XSTRNCPY(key, myKey, strlen(myKey));

    *ciphersuite = userCipher ? userCipher : "TLS13-AES128-GCM-SHA256";

    return 32;   /* length of key in octets or 0 for error */
}

int main(int argc, char** argv) {
    
    /* allocate and load config file */
    ini_t* config = ini_load("config.ini");
    int socket_port = atoi(ini_get(config, "Socket data", "socket_port"));

    /* log function related code */
    FILE *fLog = NULL;
    FILE *flogTime = NULL;
    fLog = fopen("./log/test_server.log", "a");
    flogTime = fopen("./log/test_server.exectime", "a");
    log_add_fp(fLog, LOG_TRACE);
    log_add_fp(flogTime, LOG_TIME);

    /* Vars needed by time functions */
    double w_total, cpu_total;
    w_total = cpu_total = 0;
    struct timespec w_begin, w_end;
    struct timespec cpu_begin, cpu_end;

    stopwatch_start(&w_begin, &cpu_begin);

    int listenfd, connfd;
    WOLFSSL_CTX* ctx;
    WOLFSSL* ssl;
    int n;
    char buf[MAX_LINE];
    WOLFSSL_METHOD* method;
    
    struct MemoryStruct p;
    char* acc_token = NULL;

    struct sockaddr_in address;
    int opt = 1;
    int addrlen = sizeof(address);
    char buffer[1024] = {0};

    char readKey[255];
    char readKeyID[255];

    FILE* fSec = NULL;

    fSec = fopen("./secrets.txt", "r");
    fscanf(fSec, "%s %s", readKeyID, readKey);
    fclose(fSec);

    myKeyID = (char*)malloc(sizeof(char)*(strlen(readKeyID) + 1));
    myKey = (char*)malloc(sizeof(char)*(strlen(readKey) + 1));

    strncpy(myKeyID, readKeyID, strlen(readKeyID));
    strncpy(myKey, readKey, strlen(readKey));

    myKeyID[strlen(readKeyID)] = '\0';
    myKey[strlen(readKey)] = '\0';

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Retrieved key from file and loaded in memory", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    /* Initialize wolfSSL library */
    wolfSSL_Init();

    /* Get encryption method */
    method = wolfTLSv1_3_server_method();

    /* Create wolfSSL_CTX */
    if ( (ctx = wolfSSL_CTX_new(method)) == NULL)
        log_error("wolfSSL_CTX_new error");

    wolfSSL_CTX_set_psk_server_tls13_callback(ctx, custom_psk_server_tls13_cb);
    
    //SSL_CTX_use_psk_identity_hint(ctx, "cyassl server");

    const char *defaultCipherList = "TLS13-AES128-GCM-SHA256";
    
    if (wolfSSL_CTX_set_cipher_list(ctx, defaultCipherList) != WOLFSSL_SUCCESS) {
            wolfSSL_CTX_free(ctx); ctx = NULL;
            log_error("server can't set cipher list 2");
    }

    wolfSSL_CTX_set_psk_callback_ctx(ctx, (void*)defaultCipherList);

    wolfSSL_CTX_no_dhe_psk(ctx);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Defined Wolfssl options, ready to accept connections", &w_total, &cpu_total);

    tcp_accept(&listenfd, &connfd, NULL, socket_port, 1, 0, 0, 1, 1);

    stopwatch_start(&w_begin, &cpu_begin);

    /* Create wolfSSL object */
    if ((ssl = wolfSSL_new(ctx)) == NULL)
        log_error("wolfSSL_new error");

    wolfSSL_set_fd(ssl, connfd);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Created Wolfssl object, ready to read incoming data", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);

    if ((n = wolfSSL_read(ssl, buf, (sizeof(buf) -1))) > 0) {
        stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "TLS Handshake completed, first message received", &w_total, &cpu_total);
        stopwatch_start(&w_begin, &cpu_begin);

        while ((n = wolfSSL_read(ssl, buf, (sizeof(buf) -1))) > 0) {      
            buf[n] = '\0';
            log_debug("Received from client: %s", buf);
            if (wolfSSL_write(ssl, buf, n) != n)
                err_sys("wolfSSL_write error");
        if (buf[n-1] == '9')
                break;
        }
    }
    
    if (n < 0)
        log_error("wolfSSL_read error = %d", wolfSSL_get_error(ssl, n));
    
    else if (n == 0)
        log_debug("Connection close by peer");

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Read and write completed", &w_total, &cpu_total);

    stopwatch_start(&w_begin, &cpu_begin);
    
    wolfSSL_free(ssl);
    close(connfd);
    wolfSSL_CTX_free(ctx);
    wolfSSL_Cleanup();

    free(acc_token);
    free(myKey);
    free(myKeyID);
    ini_free(config);

    stopwatch_stop(&w_begin, &w_end, &cpu_begin, &cpu_end, "Resources freed", &w_total, &cpu_total);
    log_time("Execution completed: [WALL] %.6f seconds -- [CPU] %.6f seconds\n", w_total, cpu_total);

    exit(EXIT_SUCCESS);

}
